# //
# //
# //                          AUTHOR
# //  (contact@vertexture.org)
# //  www.vertexture.org
# //  Please read on the website terms of use and licensing. Tutorials can be found also
# //
# //
# ////////////////////////////////////////////////////////////////////////////////////*/

from Qt.QtGui import *
from Qt.QtCore import *
from Qt.QtWidgets import *
import Qt.QtCompat

import maya.cmds as cmds
from pdb import set_trace as db
import maya.OpenMayaUI as apiUI

import os
import webbrowser

_winback = (38, 40, 43, 250)
_buttback = (29, 29, 35, 255)
_tabback = (16, 18, 25, 255)
_textfore = (190, 190, 193, 130)
_lineback = (16, 18, 25, 255)
_labelfor = (85, 85, 90, 100)

_gradTop = (27, 73, 135, 15)
_gradBot = (48, 147, 215, 75)

_butthov = (38, 50, 65, 130)
_frame = (235, 235, 0, 25)
_ouf = (255, 255, 0, 255)
_style = '''

QMainWindow{{
    background-color: rgba(255, 255, 25, 255);
    color : rgba(255, 255, 25, 255);
}}

QTabBar{{
    background-color: rgba{buttback};
    color : rgba{textforce};
}}

QSpinBox{{
    background-color: rgba{buttback};
    color: rgba{textforce};
}}

QGroupBox{{
    background-color: rgba{buttback};
    color: rgba{buttback};
}}

QMenuBar{{
    background-color: rgba{buttback};
}}

QMainWindow{{
    background-color: rgba{lineback};
}}

QPushButton:pressed{{
    background-color: rgba{textforce};
}}

QLabel{{
    color: rgba{labelfor};
}}

QCheckBox{{
    color: rgba{labelfor};
}}

QFrame{{
    background-color: rgba{tabback};
    color: rgba{textforce};
}}


QPushButton,QCheckBox,QComboBox{{
    background-color: rgba{buttback};
    color: rgba{textforce};
    border-radius: 1px 1px 1px 1px;
    border-color: rgb(255,10,10);
    height: 20px;
}}

QPushButton:hover, QCheckBox:hover, QComboBox:hover{{
    background-color: rgba{butthov};
    border-radius: 2px 2px 2px 2px;
}}

QPushButton:checked{{
    background-color: qlineargradient(spread:pad, x1:0.035533, y1:0, x2:0.248838, y2:1, stop:0 rgba{gradtop}, stop:1 rgba{gradbot});
    border-radius: 2px 2px 2px 2px;
}}

QLineEdit{{
    background-color: rgba{tabback};
    color: rgba{textforce};
}}

QTreeWidget:item:selected{{
    background: rgba{butthov};
    color: rgba{textforce};
}}

QTreeWidget{{
    color: rgba{textforce};
    background-color: rgba{tabback};
    show-decoration-selected: 0;
}}

;'''.format(
    winback=_winback,
    buttback=_buttback,
    tabback=_tabback,
    textforce=_textfore,
    lineback=_lineback,
    labelfor=_labelfor,
    gradtop=_gradTop,
    gradbot=_gradBot,
    butthov=_butthov,
    frame=_frame,
    ouf=_ouf)


def getMayaWin():
    winPtr = apiUI.MQtUtil.mainWindow()

    if not winPtr:
        raise Exception('could find MayaWindow Pointer')

    mayaWindow = Qt.QtCompat.wrapInstance(long(winPtr), QWidget)

    return mayaWindow


class Frameless(QMainWindow, QObject):

    prevSize = QSize()
    offset = QPoint()
    prev = QPoint()
    previous = False
    geoAttr = "geo"
    miniSize = 28

    wheeled = Signal(int)

    def __init__(self, parent=getMayaWin()):
        QMainWindow.__init__(self, parent)

        self.setContentsMargins(0, 0, 0, 0)
        self.setMouseTracking(True)
        self.setStatusBar(None)
        self.setSheet()
        self.setKey()

        self.dock = ""

    def closeEvent(self, event):
        self.settings.saveGeo(self.geometry())

        if cmds.dockControl(self.dock, exists=True):
            cmds.deleteUI(self.dock, control=True)

    def mouseMoveEvent(self, event):

        if event.buttons() == Qt.QtCore.Qt.MidButton:
            if not self.previous:
                self.prev = event.globalPos()
                self.offset = event.pos()
                self.previous = True
                return

            out = event.globalPos() - self.prev
            self.move(self.prev - self.offset + out)

        if event.buttons() == Qt.QtCore.Qt.RightButton:
            if not self.previous:
                self.prev = event.globalPos()
                self.prevSize = self.size()
                self.previous = True
                return

            out = event.globalPos() - self.prev

            self.resize(self.prevSize.width() + out.x(),
                        self.prevSize.height() + out.y()
                        )

    def mouseReleaseEvent(self, event):
        self.previous = False
        self.clicked = False
        self.on = False

    def wheelEvent(self, event):
        self.wheeled.emit(event.delta())

    def setKey(self):

        action = QAction(self)
        action.setShortcut(QKeySequence(Qt.QtCore.Qt.Key_Escape))
        action.triggered.connect(self.close)

        self.addAction(action)

    def setSheet(self):
        self.setStyleSheet(_style)

    def adddock(self, obj, title, allowed=["left", "right"], floated=False):

        self.dock = "%s%s" % (obj.objectName(), 'dock')

        if cmds.dockControl(self.dock, exists=True):
            cmds.deleteUI(self.dock, control=True)

        cmds.dockControl(
            self.dock,
            area="left",
            content=obj.objectName(),
            label=title,
            floating=floated,
            allowedArea=allowed
        )

    def addBar(self, help, toolname="", simple=False):

        self.settings = Settings(toolname)

        self.bar = QToolBar()
        self.butTheme = QPushButton('')
        self.butWindo = QPushButton('')
        self.butAbout = QPushButton('')

        self.logo = VertextureLogo(url="www.vertexture.org")
        pixlogo = QPixmap(':/vertexture/skin/vertexture/logoHs_color.png')
        self.butTheme.setIcon(self.pixmap(
            ':/vertexture/skin/bar/darkgrey.png'))
        self.butWindo.setIcon(self.pixmap(
            ':/vertexture/skin/bar/framegrey.png'))
        self.butAbout.setIcon(self.pixmap(
            ':/vertexture/skin/bar/questiongrey.png'))
        self.logo.setPixmap(pixlogo)

        # self.spacer = QWidget()
        # self.spacer = QSpacerItem(20, 20, QSizePolicy.Expanding)
        # self.spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.bar.setMinimumHeight(self.miniSize)
        self.bar.setMaximumHeight(self.miniSize)
        self.bar.setFloatable(False)
        self.bar.setMovable(False)

        self.logo.setMaximumWidth(120)

        map(lambda x: x.setMinimumHeight(self.miniSize-4), (self.butTheme,
                                                            self.butWindo,
                                                            self.butAbout))

        map(lambda x: x.setIconSize(QSize(28, 28)),
            [self.butTheme, self.butWindo, self.butAbout])

        self.butTheme.setCheckable(True)
        self.butWindo.setCheckable(True)

        self.butTheme.clicked.connect(self.themeEvent)
        self.butWindo.clicked.connect(self.windoEvent)
        self.butAbout.clicked.connect(lambda: self.aboutEvent(help))

        self.bar.addWidget(self.logo)
        # self.bar.addWidget(self.spacer)

        if not simple:
            self.bar.addWidget(self.butTheme)
            self.bar.addWidget(self.butWindo)

        self.bar.addWidget(self.butAbout)

        self.addToolBar(self.bar)
        self.addToolTips()
        self.restore()

    def addToolTips(self):
        """Description
        """
        self.butTheme.setToolTip("Change the theme to dark or light")
        self.butWindo.setToolTip("Restore/remove the window frame ")
        self.butAbout.setToolTip("bring more information about the app")

    def restore(self):
        """Description
        """
        table = {
            self.butTheme: 'theme',
            self.butWindo: "window"
        }

        for key, value in table.items():

            stored = self.settings.load(value)

            if stored:
                key.setChecked(eval(stored.capitalize()))
                key.clicked.emit()

        self.restoreGeo()

    def restoreGeo(self):
        self.setGeometry(self.settings.geo)
        self.update()

    def windoEvent(self):
        """Description
        """
        if cmds.dockControl(self.dock, exists=True):
            if not cmds.dockControl(self.dock, q=True, floating=True):
                return

        if self.sender().isChecked():
            self.setWindowFlags(Qt.QtCore.Qt.FramelessWindowHint |
                                Qt.QtCore.Qt.Window)
        else:
            self.setWindowFlags(Qt.QtCore.Qt.Window)

        self.settings.save('window', self.sender().isChecked())
        self.show()

    def themeEvent(self):
        """Description
        """
        butts = [self.butTheme, self.butWindo, self.butAbout]

        # -- darkTheme
        if self.sender().isChecked():
            self.setStyleSheet(_style)

            for butt in butts or []:
                butt.setStyleSheet("background-color: rgba(16,18,25,255)")

        # -- lightTheme
        else:

            for butt in butts or []:
                butt.setMaximumHeight(self.miniSize-2)
                butt.setFlat(True)

            self.setStyleSheet('')

        self.settings.save('theme', self.sender().isChecked())

    def pixmap(self, path):

        pix = QPixmap(path)
        pixresized = pix.scaled(self.miniSize-10, self.miniSize-10)
        return QIcon(pixresized)

    def aboutEvent(self, help):
        """Description
        """
        about = AboutWin(help)
        about.size(400, 800)
        about.setWindowTitle("Manual Reference")
        about.setWindowIcon(
            QIcon(QPixmap(':/vertexture/skin/vertexture/logoHs_color.png')))
        about.show()

# ====================================================================
# ====================================================================


class OpenGl(Frameless, QObject):

    def __init__(self, parent=getMayaWin()):
        QMainWindow.__init__(self, parent)
        pass


# ====================================================================
# ====================================================================
# ====================================================================
# ====================================================================

class Confirmation(QWidget, QObject):

    _output = str()

    returned = Signal(tuple)

    _result = str()

    def __init__(self, genre="path",
                 label="",
                 infos="",
                 boomrang="",
                 placeHolder="",
                 object=None,
                 parent=None):

        QWidget.__init__(self, parent)

        self.placeHolder = placeHolder
        self.boomrang = boomrang
        self.object = object
        self.label = label
        self.infos = infos
        self.genre = genre
        self.setWin()

        self.setPrefs()
        self.setConnections()

        self.mainPop.show()

    def setConnections(self):
        self.okButton.clicked.connect(self.apply)
        self.cancelButton.clicked.connect(self.ignore)
        self.input.returnPressed.connect(self.apply)
        self.input.installEventFilter(self)

    def eventFilter(self, obj, event):

        if obj == self.input:
            if (event.type() == QEvent.KeyPress or
                    event.type() == QEvent.KeyRelease):

                self.checkInput()
                obj.event(event)
                return True
        return False

    def __repr__(self):
        return self._result

    def checkInput(self):
        self.log.clear()

        text = str(self.input.text())

        if self.genre == "path":
            if not os.path.exists(text):
                self.okButton.setHidden(True)
                self.log.setText("\'{}\' does not exist".format(text))

            else:
                self.log.setText("status ok".format())
                self.okButton.setHidden(False)

    def ignore(self):

        self.returned.emit(tuple())
        self.mainPop.close()

    def apply(self):
        self.checkInput()
        path = str(self.input.text())
        self.returned.emit((path, self.boomrang))

        self._result = path
        self.mainPop.close()

    def setPrefs(self):
        self.notice.setAlignment(
            Qt.QtCore.Qt.AlignHCenter | Qt.QtCore.Qt.AlignVCenter)
        self.log.setAlignment(Qt.QtCore.Qt.AlignHCenter |
                              Qt.QtCore.Qt.AlignVCenter)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.okButton.setHidden(True)

        self.mainPop.setWindowModality(Qt.QtCore.Qt.WindowModal)

        if self.placeHolder:
            self.input.setText(self.placeHolder)
            self.checkInput()

    def setWin(self):

        self.mainPop = Frameless()

        self.notice = QLabel(self.label)
        self.input = QLineEdit("")
        self.log = QLabel(self.infos)

        self.okButton = QPushButton("Confirm", self)
        self.cancelButton = QPushButton("Ignore", self)

        mainLayout = QVBoxLayout(self)
        self.vlayout = QVBoxLayout(self)
        self.hlayout = QHBoxLayout(self)

        self.vlayout.addWidget(self.notice)
        self.vlayout.addWidget(self.input)
        self.vlayout.addWidget(self.log)

        self.hlayout.addWidget(self.okButton)
        self.hlayout.addWidget(self.cancelButton)

        self.setLayout(mainLayout)
        mainLayout.addLayout(self.vlayout)
        mainLayout.addLayout(self.hlayout)

        self.mainPop.setCentralWidget(self)


class Prefs(QWidget, QObject):

    prefsChanged = Signal(str)

    def __init__(self):
        QWidget.__init__(self, parent)

        self.setWin()
        self.show()

    def setWin(self):

        self.mainPop = Frameless()


class ZiInvertButton(QWidget, QObject):
    """A inverted display label, with a clicked signal
    """

    clicked = Signal()

    def __init__(self, label='Title', parent=None):
        QWidget.__init__(self, parent)

        self.setMouseTracking(True)

        self._label = label
        self._border = 5
        self._light = 170
        self._on = False
        self.hovered = False

        self._fillColor = QColor(126, 131, 90)

    @property
    def light(self):
        return self._light

    @light.setter
    def light(self, value):
        self._light = value
        self.update()

    @property
    def label(self):
        return str(self._label)

    @label.setter
    def label(self, value):
        self._label = value
        self.update()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        self.update()

    @property
    def on(self):
        return self._on

    @on.setter
    def on(self, value):
        self._on = value
        self.update()

    @property
    def fillColor(self):
        return self._fillColor

    @fillColor.setter
    def fillColor(self, value):
        self._fillColor = value
        self.update()

    def mouseMoveEvent(self, event):

        if QRectF(0, 0,
                  self.size().width(),
                  self.size().height()).contains(event.pos()):

            self.hovered = True

        else:
            self.hovered = False

        self.repaint()

    def leaveEvent(self, event):
        self.hovered = False
        self.update()

    def mousePressEvent(self, event):

        if event.button() == Qt.QtCore.Qt.LeftButton:
            self.on = False if self.on is True else True
            self.clicked.emit()
            self.update()

    def paintEvent(self, event):

        painter = QPainter(self)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)

        font = QFont('MS Sans Serif', 8, QFont.Light)

        borderBrush = QBrush(QColor(200, 200, 200, 255))

        textPath = QPainterPath()
        textPath.addText(0, 0, font, self.label)

        bound = textPath.boundingRect()
        w = self.size().width()
        h = self.size().height()
        margingx = (w - bound.width()) * .5
        margingy = (h - bound.height()) * .5
        textPath.translate(margingx, margingy + bound.height())

        # -- border
        if self.hovered:
            painter.setPen(borderBrush.color())
        else:
            painter.setPen(QColor(77, 77, 77))

        if self.on:
            textBrush = self.palette().window()
            painter.setBrush(QBrush(self.fillColor))
        else:
            textBrush = QBrush(QColor(200, 200, 200, 255))
            painter.setBrush(QBrush(Qt.QtCore.Qt.NoBrush))

        painter.drawRoundedRect(0, 0, w - 1, h - 1, self.border, self.border)

        # -- text
        painter.setBrush(textBrush)
        painter.setPen(textBrush.color())
        painter.drawPath(textPath)

        painter.end()


class InteractiveText(QTextEdit, QObject):
    """Send a dblClick signal with the text double clicked
    """

    rightClick = Signal()
    dblClick = Signal(str)
    midClick = Signal()

    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)

    def mouseDoubleClickEvent(self, event):
        QTextEdit.mouseDoubleClickEvent(self, event)

        cursor = self.textCursor()

        if cursor.hasSelection():
            text = cursor.selectedText()

            self.dblClick.emit(text)

    def mousePressEvent(self, event):

        if event.buttons() == Qt.QtCore.Qt.MidButton:
            self.midClick.emit()

        if event.buttons() == Qt.QtCore.Qt.RightButton:
            self.rightClick.emit()


class AboutWin(QWidget, QObject):
    """Display a widown with html
    """

    def __init__(self, text, parent=None):
        super(AboutWin, self).__init__(parent)

        self.main = Frameless()
        self.doc = InteractiveText()
        self.doc.document().setDefaultStyleSheet(
            "p,li { white-space: pre-wrap; }")

        self.doc.setHtml(text)
        self.doc.setReadOnly(True)

        self.lay = QVBoxLayout()
        self.lay.addWidget(self.doc)
        self.setLayout(self.lay)

        self.main.setCentralWidget(self)
        self.main.show()

    def setContent(self, txt):
        self.doc.setHtml(txt)

    def setWindowTitle(self, title):
        """Set the title of the windown
        """
        self.main.setWindowTitle(title)

    def size(self, x=470, y=170):
        """Reset the size of the window
        """
        self.main.resize(x, y)


class MimeWin(QWidget, QObject):

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.setAcceptDrops(True)
        self.resize(500, 500)

    def mousePressEvent(self, event):
        if event.button() == Qt.QtCore.Qt.LeftButton:
            print 'in Mime %s' % mime.text()

    def dragEnterEvent(self, event):

        if event.mouseButtons() == Qt.QtCore.Qt.MidButton:
            event.acceptProposedAction()
            print 'dragEnterEvent accpeted'

    def dragLeaveEvent(self, event):
        pass

    def dragMoveEvent(self, event):
        pass

    def dropEvent(self, event):
        # print 'dropEvent'

        mime = QMimeData()
        mime.setText(';'.join(cmds.ls(sl=True)))

        print 'in Mime %s' % mime.text()

# class ziPanel(QFrame, QObject):
#     def __init__(self, title="", parent=None):
#         QFrame.__init__(self, parent)
#         self._title = title
#         self.border = 3
#         self.light = 150

#     def paintEvent(self, event):

#         painter = QPainter(self)
#         painter.setRenderHint(QPainter.Antialiasing, True)

#         # textBrush = self.palette().base()
#         textColor = QColor(255, 100, 100)
#         font = QFont('FreeSans', 12, QFont.Bold)


#         # painter.setPen(Qt.QtCore.Qt.NoPen)
#         # painter.setBrush(QBrush(textColor.lighter(self.light)))

#         painter.setBrush(QBrush(QColor(textColor)))
#         # textLabel = QPainterPath()
#         # textLabel.addText(5, 5, font, self._title)

#         # bound = textLabel.boundingRect()
#         painter.drawRoundedRect(0, 0, 200, 200, self.border, self.border)

#         # painter.setBrush(textBrush)
#         # painter.drawPath(textLabel)

#         painter.end()

class ziFrame(QFrame, QObject):

    def __init__(self, parent=None):
        QFrame.__init__(self, parent)

        self.pixup = QPixmap(":/vertexture/skin/vertexture/layer/arrow_up.png")
        self.pixdn = QPixmap(":/vertexture/skin/vertexture/layer/arrow_dn.png")
        self.collapeBtn = QPushButton()
        self.collapeBtn.setIcon(QIcon(self.pixdn))

        self.setWidget()

    def setWidget(self):

        self.hlayout = QHBoxLayout()
        self.addWidget(self.collapeBtn)

    def addWidget(self, widget):
        self.layout().addWidget(widget)

    # def paintEvent(self, event):
    #     painter = QPainter(self)
    #     painter.setRenderHint(QPainter.HighQualityAntialiasing)

    #     self.w = self.size().width()
    #     self.h = self.size().height()

    #     radius = 2
    #     position = QRect(self.w - 35, 5, 20, 20)

    #     painter.setPen(QPen(QColor(255, 255, 0), 3))
    #     painter.drawPixmap(position, self.pixup)
    #     painter.drawText(self.w * .5, self.h * .5, "dedede")

    #     # painter.drawRoundedRect(self.w - 35, 5, 30, 20, radius, radius, )
    #     painter.end()


class VertextureLogo(QLabel, QObject):

    def __init__(self, url, parent=None):
        QLabel.__init__(self, parent)
        self.url = url
        # self.setCursor(Qt.QtCore.Qt.DragMoveCursor)

    def mousePressEvent(self, event):
        # if event.button() == Qt.QtCore.Qt.LeftButton:
            # webbrowser.open(self.url, new=2)
        pass


class Settings(QObject):

    def __init__(self, name):
        self.qsettings = QSettings('Vertexture', name)

    def clear(self):
        self.qsettings.clear()

    def save(self, attr, value):
        self.qsettings.setValue(attr, value)

    def saveGeo(self, geo):

        self.save("x", geo.x())
        self.save("y", geo.y())
        self.save("width", geo.width())
        self.save("height", geo.height())

    @property
    def geo(self):

        x = self.load("x")
        y = self.load("y")
        h = self.load("width")
        w = self.load("height")

        if x and y and h and w:
            return QRect(x, y, h, w)
        else:
            return QRect(0, 0, 500, 500)

    def load(self, attr):

        if self.qsettings.contains(attr):
            return self.qsettings.value(attr)


class ziScrub(QWidget, QObject):
    """Text displaying different value with a move drag
    """

    scrubed = Signal(object)
    family = 'MS Sans Serif'
    fuzzed = Signal(str)
    hovered = Signal(str)
    changed = Signal(float)

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.setFocusPolicy(Qt.QtCore.Qt.StrongFocus)
        self.setMouseTracking(True)

        self.index = 0  # -- to del for families

        self.mainPath = QPainterPath()
        self.text = 'Property'
        self.weight = 'light'
        self._value = 0

        self._min = 0
        self._max = 100

        self.textColor = QColor(255, 255, 255, 255).darker(140)
        self.wasFloat = False
        self.clicked = False
        self.openKey = False
        self.waiting = False
        self.fontSize = 10
        self._data = False
        self.count = int()
        self.prev = False
        self.on = False
        self.step = 1
        self.raw = 0

        self.set_type("int")
        self.font = QFont()
        self.sign = 0

        self.setMinimumHeight(20)

    def set_type(self, typ):

        if typ == 'int':
            self.int, self.float, self.string = [True, False, False]

        if typ == 'float':
            self.int, self.float, self.string = [False, True, False]

        if typ == 'string':
            self.int, self.float, self.string = [False, False, True]

    def set_behaviors(self,
                      text,
                      max,
                      min,
                      value,
                      typ="int",
                      weight="normal",
                      fontsize=10
                      ):

        self.max = max
        self.min = min
        self.text = text
        self.value = value
        self.value2raw(value)
        self.set_type(typ)
        self.weight = weight

    @classmethod
    def setSize(cls, value):
        cls.fontSize = value

    @classmethod
    def setFont(cls, value):
        cls.family = value

    @classmethod
    def setHeight(cls, value):
        cls.setMaximumHeight(value)

    @property
    def min(self):
        return self._min

    @min.setter
    def min(self, min):
        self._min = min

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, max):
        self._max = max

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        self.update()

    @property
    def fontSize(self):
        return self._fontSize

    @fontSize.setter
    def fontSize(self, value):
        self._fontSize = value
        self.update()

    @property
    def w(self):
        return self.size().width()

    @property
    def h(self):
        return self.size().height()

    @property
    def textColor(self):
        return self._txtcolor

    @textColor.setter
    def textColor(self, value):
        self._txtcolor = value

    @property
    def step(self):
        return self._inc

    @step.setter
    def step(self, value):
        self._inc = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self.int, self.float, self.string = [False, False, False]

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, value):

        if value == 'light':
            self._weight = QFont.Light

        if value == 'normal':
            self._weight = QFont.Normal

        if value == 'demiBold':
            self._weight = QFont.DemiBold

        if value == 'bold':
            self._weight = QFont.Bold

        if value == 'black':
            self._weight = QFont.Black

        self.update()

    @property
    def value(self):
        return self._value

    def raw2value(self, raw):
        res = self.min + ((raw / 100.0) * (self.max - (self.min)))
        self._value = self.max if res > self.max else res
        self._value = self.min if res < self.min else res

        self.value = res

    def value2raw(self, value):
        self.raw = (value - self.min) / float((self.max - (self.min))) * 100
        self.repaint()

    @value.setter
    def value(self, value):

        if self.string:
            self._value = str(value)
            self.update()
            return

        if self.data:
            self._value = value

            self.update()
            return

        # -- if specific datas (i.e 8,16,32)
        else:
            if self._data:

                if value in self._data:
                    ind = self.data.index(value)

                    if self.value > value:
                        self._value = self.data[ind - 1]
                        return

                    if self.value < value:
                        if len(self.data) > ind + 1:
                            self._value = self.data[ind + 1]
                            return
            elif value == '':
                self._value = value
                return

            # -- int or float
            else:
                self._value = value

        self.changed.emit(self._value)
        self.value2raw(value)
        self.update()

    def mousePressEvent(self, event):

        if event.buttons() == Qt.QtCore.Qt.LeftButton:
            self.clicked = True
            self.prev = False

            self.openKey = True
            self.waiting = True

            # self._value = '_'

            self.update()
            event.accept()

    def keyPressEvent(self, event):

        if event.key() == Qt.QtCore.Qt.Key_Enter \
                or event.key() == Qt.QtCore.Qt.Key_Escape \
                or event.key() == Qt.QtCore.Qt.Key_Return:
            self.openKey = False

            if self.wasFloat:
                self.wasFloat = False
                self.set_type('float')
                self.value = round(float(self.value), 2)

            self.scrubed.emit(self.value)

            self.update()
            event.accept()
            return

        # -- correcting by deleting previous keystroke
        if event.key() == Qt.QtCore.Qt.Key_Backspace:
            if isinstance(self.value, str):
                self.value = self.value[:-1]
                if len(self.value) < 2:
                    self.value = ''

            elif isinstance(self.value, int):
                self.value = int(str(self.value)[:-1])

            elif isinstance(self.value, float):
                self.value = float(str(self.value)[:-1])

            self.update()
            event.accept()
            return

        # -- typing
        if self.openKey:

            if self.value == '_':
                try:
                    self.value = str(event.text()) if self.string\
                        else int(event.text())
                except ValueError:
                    pass
            else:
                newValue = '{}{}'.format(self.value, event.text())

                try:
                    if self.wasFloat:
                        self.set_type('float')
                    if self.string:
                        self.value = str(newValue)
                    if self.int:
                        self.value = int(newValue)
                    if self.float:
                        if event.text() == '.':
                            self.set_type('string')
                            self.value = str(newValue)
                            self.wasFloat = True
                        else:
                            self.value = float(newValue)
                except ValueError:
                    pass

                self.fuzzed.emit(str(self.value))

            self.update()
            event.accept()

    def mouseDoubleClickEvent(self, event):

        if self.data:
            return

    def mouseMoveEvent(self, event):

        if self.mainPath.boundingRect().contains(event.pos()):
            self.hovered.emit(self.toolTip())
            self.on = True
        else:
            self.on = False

        if self.clicked:

            current = event.pos()
            self.raw = current.x() / float(self.w) * 100

            self.raw = 100 if self.raw > 100 else self.raw
            self.raw = 0 if self.raw < 0 else self.raw

            self.raw2value(self.raw)

        self.repaint()

    def leaveEvent(self, event):
        self.on = False
        self.update()

    def mouseReleaseEvent(self, event):

        self.clicked = False
        self.sign = 0
        self.update()

    def paintEvent(self, event):

        painter = QPainter(self)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)
        self.mainPath = QPainterPath()

        # 0 --->> w
        xpos = self.raw / float(100 - 0) * (self.w)

        xpos = self.w if xpos > self.w else xpos
        xpos = 0 if xpos < 0 else xpos

        # -- jauge
        if self.on:

            gradJauge = QLinearGradient(0, 0, xpos, self.h)

            coloJauge = QColor(126, 131, 90, 255)

            coloJauge.setAlpha(0)
            gradJauge.setColorAt(0.3, coloJauge)
            coloJauge.setAlpha(255)
            gradJauge.setColorAt(1, coloJauge)

            painter.setPen(QPen(Qt.QtCore.Qt.NoPen))
            painter.setBrush(QBrush(gradJauge))
            painter.drawRoundedRect(4, self.h * .2,
                                    xpos - 4,
                                    self.h * .6, 3, 3)

            painter.setBrush(QBrush(Qt.QtCore.Qt.NoBrush))

        # -- bottom line
        gradient = QLinearGradient(0, 0, self.w, self.h)
        gradColor = self.textColor
        blankColor = QColor(self.textColor).darker(290)

        if self.on:
            gradColor = self.textColor.lighter(190)
            blankColor = blankColor.lighter(190)

        if self.sign > 0:
            gradColor = QColor(gradColor.red(), 0, 0)

        if self.sign < 0:
            gradColor = QColor(0, 0, gradColor.blue())

        if self.sign == 0:
            gradColor = QColor(gradColor.red(),
                               gradColor.green(),
                               gradColor.blue()
                               )

        gradColor.setAlpha(0)
        gradient.setColorAt(0, gradColor)

        gradColor.setAlpha(255)
        gradient.setColorAt(0.35, gradColor)
        gradient.setColorAt(0.65, gradColor)

        gradColor.setAlpha(0)
        gradient.setColorAt(1, gradColor)
        gradColor.setAlpha(255)

        painter.setPen(QPen(gradient, 1.0))
        painter.drawLine(0, self.h * .8, self.w, self.h * .8)

        # -- border
        painter.setPen(QPen(blankColor, 1.0))
        self.mainPath.addRoundedRect(0, 0, self.w - 1, self.h - 1, 2, 2)
        painter.drawPath(self.mainPath)

        # -- text
        self.font.setPointSize(self.fontSize)
        self.font.setWeight(self.weight)
        self.font.setFamily(self.family)
        painter.setFont(self.font)

        painter.setPen(QPen(self.textColor, 1.0))

        value = int(self.value) if self.int is True else "{:.2f}".format(
            float(self.value))

        painter.drawText(5, self.h * .7, "{} {}".format(self.text, value))

        # -- knob
        painter.setPen(QPen(Qt.QtCore.Qt.NoPen))
        knobColor = QColor(self.textColor)

        knobColor.setAlpha(60)
        painter.setBrush(knobColor)

        radius = 12
        painter.drawEllipse(xpos - (radius * .5),
                            self.h * .20,
                            radius, radius
                            )

        knobColor.setAlpha(255)
        painter.setBrush(knobColor)
        radius = radius * .6
        painter.drawEllipse(xpos - (radius * .5),
                            (self.h * .20) + radius * .5,
                            radius,
                            radius
                            )

        painter.end()


class ZiLayer(QWidget, QObject):
    """Photoshop like layer widget
    display a label with icons

    usage:
    lay = Layer()
    lay.setPixFirst(pixmap, pixmap)
    lay.setPixSecond(pixmap, pixmap)
    lay.setThumb(pixmap)
    lay.setLabel(string)
    """

    DBLClicked = Signal(str)
    MMBClicked = Signal()
    LMBClicked = Signal()
    RMBClicked = Signal(str)

    firstClicked = Signal()
    secondClicked = Signal()
    thumbClicked = Signal()
    # labelClicked= Signal(str)

    clickedColor = outColor = QColor(39, 40, 46)

    def __init__(self, label, opacity, mixes, option="", parent=None):
        QWidget.__init__(self, parent)

        self.addPath = QPainterPath()
        self.firstOff = QPixmap()
        self.firstOn = QPixmap()

        self.secondOff = QPixmap()
        self.secondOn = QPixmap()

        self.thumbOff = QPixmap()
        self.thumbOn = QPixmap()

        self._selected = False
        self._thumbed = False
        self.hovered = False

        self.thumb = QPixmap()
        self.firstPix = QPixmap()
        self.secondPix = QPixmap()
        self.rect = QRect(80, 0, self.size().width() - 50,
                          self.size().height())

        self.setMaximumSize(950, 50)
        self.setMinimumSize(100, 40)
        self.insideColor = QColor(61, 62, 70)

        self._label = label
        self._option = option
        self._font = QFont()
        self.setFont()

        # self.scrub = ""
        self.setMouseTracking(True)

    # def setScrub(self, scrub):
        # self.scrub = scrub

    def set_behaviors(self, label="layer1", option=""):

        self.setLabel(label)
        self.option = option

    def setPixFirst(self, openPix, closePix, size=20):

        self.firstOff = closePix.scaled(size, size)
        self.firstOn = openPix.scaled(size, size)

        self.firstPix = self.firstOn

    def setPixSecond(self, openPix, closePix, size=20):

        self.secondOff = closePix.scaled(size, size)
        self.secondOn = openPix.scaled(size, size)

        self.secondPix = self.secondOn

    def setPixThumb(self, onPix, offPix, size=20):

        self.thumbOff = offPix.scaled(size, size)
        self.thumbOn = onPix.scaled(size, size)

        self.thumb = self.thumbOff

    def setLabel(self, label):
        self._label = label

    def swapSecond(self, invert=True, switch='ON'):

        if invert:
            self.secondPix = self.secondOff if self.secondPix == self.secondOn else self.secondOn

        else:
            if switch == 'ON':
                self.secondPix = self.secondOff

            if switch == 'OFF':
                self.secondPix = self.secondOn

        self.update()

    def swapFirst(self, invert=True, switch='ON'):

        if invert:
            self.firstPix = self.firstOff if self.firstPix == self.firstOn else self.firstOn

        else:
            if switch == 'ON':
                self.firstPix = self.firstOff

            if switch == 'OFF':
                self.firstPix = self.firstOn

        self.update()

    def swapThumb(self, invert=True, switch='ON'):

        if invert:
            self.thumb = self.thumbOff if self.thumb == self.thumbOn else self.thumbOn
            self.thumbed = False if self.thumb == self.thumbOff else True

        else:
            if switch == 'ON':
                self.thumb = self.thumbOn
                self.thumbed = True

            if switch == 'OFF':
                self.thumb = self.thumbOff
                self.thumbed = False

        self.update()

    @property
    def insideColor(self):
        return self._incolor

    @insideColor.setter
    def insideColor(self, color):
        self._incolor = color

    @property
    def option(self):
        if self._option == '()':
            return ''
        return self._option

    @property
    def selected(self):
        return self._selected

    @property
    def thumbed(self):
        return self._thumbed

    @property
    def label(self):
        if self._label:
            return str(self._label)
        else:
            return ''

    @property
    def font(self):
        return self._font

    @property
    def secondIsOn(self):
        value = False if self.secondPix == self.secondOn else True
        return value

    @property
    def firstIsOn(self):
        value = False if self.firstPix == self.firstOff else True
        return value

    @selected.setter
    def selected(self, value):
        self._selected = value
        self.update()

    @property
    def file(self):
        return self._file

    @label.setter
    def label(self, value):
        self._thumbed = value
        self.update()

    @thumbed.setter
    def thumbed(self, value):
        self._thumbed = value
        self.update()

    @option.setter
    def option(self, value):
        self._option = str(value)

    def setFont(self, size=11, weight=25, family="MS Sans Serif"):
        self._font.setPointSize(size)
        self._font.setWeight(weight)
        self._font.setFamily(family)
        self.update()

    def leaveEvent(self, event):
        self.hovered = False
        self.clickedColor = self.outColor
        self.update()

    def mouseDoubleClickEvent(self, event):
        self.DBLClicked.emit(self.label)

    def mouseMoveEvent(self, event):

        if self.rect.contains(event.pos()):
            self.hovered = True
            self.clickedColor = self.insideColor

        else:
            self.hovered = False
            self.clickedColor = self.outColor

        self.update()
        return

    def mousePressEvent(self, event):

        if event.buttons() == Qt.QtCore.Qt.LeftButton:
            if self.rectFirst.contains(event.pos()):
                self.swapFirst()
                self.firstClicked.emit()

            if self.rectSecond.contains(event.pos()):
                self.swapSecond()
                self.secondClicked.emit()

            if self.rectTmb.contains(event.pos()):
                self.swapThumb()
                self.thumbClicked.emit()

            if self.rect.contains(event.pos()):
                self._selected = not self._selected
                self.LMBClicked.emit()

            self.update()

        if event.buttons() == Qt.QtCore.Qt.RightButton:
            if self.rectFirst.contains(event.pos()):
                self.RMBClicked.emit('first')
                return

            if self.rectSecond.contains(event.pos()):
                self.RMBClicked.emit('second')
                return

            if self.rectTmb.contains(event.pos()):
                self.RMBClicked.emit('thumb')
                return

            if self.rect.contains(event.pos()):
                self.RMBClicked.emit('main')
                return

        if event.buttons() == Qt.QtCore.Qt.MidButton:
            print 'middle mouse'
            self.MMBClicked.emit()

    def paintEvent(self, event):

        painter = QPainter(self)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)

        border = 4
        thumSiz = 22
        icoSiz = thumSiz - border
        size = self.size()

        backgGrad = QLinearGradient()
        color = QColor(50, 50, 50, 100)
        avar = 1.5 if self.selected else 1

        backgGrad.setColorAt(0.00, color.lighter(0 * avar))
        backgGrad.setColorAt(0.45, color.lighter(102 * avar))
        backgGrad.setColorAt(0.5, color.lighter(40 * avar))
        backgGrad.setColorAt(1.0, color.lighter(60 * avar))

        backgGrad.setStart(QPointF(0, 0))
        backgGrad.setFinalStop(QPointF(0, self.height()))

        # ------------------------------------ BACKGROUND
        painter.setBrush(backgGrad)
        painter.drawRoundedRect(0, 0,
                                size.width() - border,
                                size.height() - border,
                                border, border)

        # ------------------------------------ BORDER
        painter.setBrush(QBrush(Qt.QtCore.Qt.NoBrush))

        gradient = QLinearGradient(0, 0, size.width(), 0)
        gradColor = QColor(self.insideColor)

        gradColor.setAlpha(200)
        gradient.setColorAt(0, gradColor)

        gradColor.setAlpha(0)
        gradient.setColorAt(0.35, gradColor)
        gradient.setColorAt(0.65, gradColor)

        gradColor.setAlpha(200)
        gradient.setColorAt(1, gradColor)

        if not self.hovered:
            painter.setPen(QPen(QBrush(gradient), 1))
        if self.hovered:
            painter.setPen(QPen(QBrush(QColor(180, 180, 180)), 1))
        if self.selected:
            painter.setPen(QPen(QBrush(QColor(126, 131, 90)), 1))

        painter.drawRoundedRect(0, 0,
                                size.width() - border,
                                size.height() - border,
                                border, border)

        # ------------------------------------ FIRST
        self.rectFirst = QRect(border * 2, border, icoSiz, icoSiz)

        if not self.firstPix.isNull():
            painter.drawPixmap(self.rectFirst, self.firstPix)

        # ------------------------------------ SECOND
        self.rectSecond = QRect(self.rectFirst.right() + border,
                                border,
                                icoSiz,
                                icoSiz)

        if not self.secondPix.isNull():
            painter.drawPixmap(self.rectSecond, self.secondPix)

        # ------------------------------------ THUMB
        self.rectTmb = QRect(self.rectSecond.right() + (icoSiz * .7),
                             border,
                             thumSiz,
                             thumSiz)

        painter.drawPixmap(self.rectTmb, self.thumb)

        if not self.hovered:
            painter.setBrush(QBrush(QColor(153, 153, 155, 200)))

        if self.hovered or self.selected:
            painter.setBrush(QBrush(QColor(233, 233, 235, 255)))

        painter.setBrush(QBrush(Qt.QtCore.Qt.NoBrush))
        painter.setFont(self.font)
        painter.setPen(QPen(QColor(153, 153, 155, 255), .5))
        # ------------------------------------ TEXT
        rectText = QRect(self.rectTmb.right() + border,
                         self.rectTmb.bottom(),
                         icoSiz,
                         icoSiz)

        painter.drawText(rectText.left(),
                         self.rectTmb.bottom() - self.rectTmb.top(),
                         self.label)

        textPath = QPainterPath()
        textPath.addText(rectText.left(),
                         self.rectTmb.bottom() - self.rectTmb.top(),
                         self.font,
                         self.label)

        # ----------------------------------- OPTION >> text(option)
        optionPath = QPainterPath()
        painter.setFont(QFont(self.font.family(),
                              int(self.font.pointSize() * .5)))

        painter.drawText(textPath.boundingRect().right() + (icoSiz * .5),
                         self.rectTmb.bottom() - self.rectTmb.top(),
                         self.option)
